import logging
import warnings


def import_pytorch_lightning():
    """
    Import pytorch_lightning with warnings suppressed.
    """
    # filter deprecation warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import pytorch_lightning as pl

    warnings.filterwarnings("default", category=DeprecationWarning)
    warnings.filterwarnings("ignore", ".*does not have many workers.*")
    warnings.filterwarnings("ignore", ".*GPU available but not used.*")
    logger = logging.getLogger("pytorch_lightning.utilities.rank_zero")
    logger.setLevel(logging.ERROR)
    return pl



import unittest
from functools import lru_cache

import numpy as np

import neurosym as ns

arith_dsl = ns.examples.mutable_arith_combinators_dsl

out_t = ns.parse_type("(i) -> i")


@lru_cache(maxsize=None)
def corpus():
    fam = ns.BigramProgramDistributionFamily(arith_dsl, valid_root_types=[out_t])
    dist = fam.uniform()
    return sorted(
        {
            fam.sample(dist, np.random.RandomState(i), depth_limit=20)
            for i in range(100)
        },
        key=str,
    )


class BasicDSLTest(unittest.TestCase):
    def assertDSL(self, dsl, expected):
        dsl = "\n".join(
            sorted([line.strip() for line in dsl.split("\n") if line.strip()])
        )
        expected = "\n".join(
            sorted([line.strip() for line in expected.split("\n") if line.strip()])
        )
        print(dsl)
        self.maxDiff = None
        self.assertEqual(dsl, expected)

    def test_basic_expand(self):
        self.assertDSL(
            arith_dsl.render(),
            """
            * :: (i -> i, i -> i) -> i -> i
            + :: (i -> i, i -> i) -> i -> i
            1 :: () -> i -> i
            count[counter] :: () -> i -> i
            even? :: (i -> i) -> i -> i
            ite :: (i -> i, i -> i, i -> i) -> i -> i
            x :: () -> i -> i
            """,
        )

    def test_independent_mutability(self):
        prog = ns.parse_s_expression("(ite (even? (x)) (count) (count))")
        fn_1 = arith_dsl.compute(arith_dsl.initialize(prog))
        fn_2 = arith_dsl.compute(arith_dsl.initialize(prog))
        self.assertEqual([fn_1(2), fn_1(4), fn_1(8)], [1, 2, 3])
        self.assertEqual([fn_1(2), fn_1(4), fn_1(8)], [4, 5, 6])
        # fn_2 is independent of fn_1
        self.assertEqual([fn_2(2), fn_2(4), fn_2(8)], [1, 2, 3])
        self.assertEqual([fn_1(1), fn_1(2), fn_1(3)], [1, 7, 2])
        self.assertEqual([fn_1(1), fn_1(2), fn_1(3)], [3, 8, 4])


class CompressionTest(unittest.TestCase):
    def fuzzy_check_fn_same(self, fn_1, fn_2):
        inputs = np.random.RandomState(0).randint(100, size=100)
        outputs_1 = [fn_1(x) for x in inputs]
        outputs_2 = [fn_2(x) for x in inputs]
        self.assertEqual(outputs_1, outputs_2)

    def test_single_step(self):
        dsl2, rewritten = ns.compression.single_step_compression(arith_dsl, corpus())
        self.assertEqual(len(rewritten), len(corpus()))
        for orig, rewr in zip(corpus(), rewritten):
            self.fuzzy_check_fn_same(
                arith_dsl.compute(arith_dsl.initialize(orig)),
                dsl2.compute(dsl2.initialize(rewr)),
            )

    def test_multi_step(self):
        dsl2, rewritten = ns.compression.multi_step_compression(arith_dsl, corpus(), 5)
        self.assertEqual(len(rewritten), len(corpus()))
        for orig, rewr in zip(corpus(), rewritten):
            self.fuzzy_check_fn_same(
                arith_dsl.compute(arith_dsl.initialize(orig)),
                dsl2.compute(dsl2.initialize(rewr)),
            )


class BasicProcessDSL(unittest.TestCase):
    def assertDSL(self, dsl, expected):
        dsl = "\n".join(
            sorted([line.strip() for line in dsl.split("\n") if line.strip()])
        )
        expected = "\n".join(
            sorted([line.strip() for line in expected.split("\n") if line.strip()])
        )
        print(dsl)
        self.maxDiff = None
        self.assertEqual(dsl, expected)

    def test_basic_expand(self):
        self.assertDSL(
            ns.examples.basic_arith_dsl(True).render(),
            """
            $0_0 :: V<i@0>
            $1_0 :: V<i@1>
            $2_0 :: V<i@2>
            + :: (i, i) -> i
            1 :: () -> i
            lam_0 :: L<#body|i;i> -> (i, i) -> #body
            lam_1 :: L<#body|i> -> i -> #body
            """,
        )

    def setUp(self):
        self.dsl = ns.examples.basic_arith_dsl(True)
        print(ns.examples.basic_arith_dsl(True).render())
        self.fn_code = "(lam_0 (+ ($1_0) ($0_0)))"

    def test_compute(self):
        fn = self.dsl.compute(self.dsl.initialize(ns.parse_s_expression(self.fn_code)))
        self.assertEqual(fn(1, 2), 3)
        self.assertEqual(fn(3, 4), 7)
        self.assertEqual(fn(1000, -24), 976)

    def test_basic_compress(self):
        code = [
            "(lam_0 (+ (+ (1) (1)) ($0_0)))",
            "(lam_0 (+ (1) ($0_0)))",
        ]
        code = [ns.parse_s_expression(x) for x in code]
        dsl2, rewritten = ns.compression.single_step_compression(self.dsl, code)
        self.assertEqual(len(rewritten), len(code))
        self.assertEqual(
            [ns.render_s_expression(x) for x in rewritten],
            [
                "(__10 (+ (1) (1)))",
                "(__10 (1))",
            ],
        )
        self.assertEqual(
            dsl2.productions[-1].render().strip(),
            "__10 :: i -> (i, i) -> i = (lam-abstr (#0) (lam_0 (+ #0 ($0_0))))",
        )

    def test_multi_argument(self):
        code = [
            "(lam_0 (+ (1) ($1_0)))",
            "(lam_0 (+ (1) (2)))",
        ]
        code = [ns.parse_s_expression(x) for x in code]
        dsl2, rewritten = ns.compression.single_step_compression(self.dsl, code)
        self.assertEqual(len(dsl2.productions), len(self.dsl.productions))
        self.assertEqual(rewritten, code)

    def test_compress_yoinking_variables(self):
        code = [
            "(lam_1 (lam_1 (+ (1) ($1_0))))",
            "(lam_1 (+ (1) (2)))",
        ]
        code = [ns.parse_s_expression(x) for x in code]
        dsl2, rewritten = ns.compression.single_step_compression(self.dsl, code)
        self.assertEqual(
            dsl2.productions[-1].render().strip(),
            "__10 :: i -> i -> i = (lam-abstr (#0) (lam_1 (+ (1) #0)))",
        )
        self.assertEqual(len(rewritten), len(code))
        self.assertEqual(
            [ns.render_s_expression(x) for x in rewritten],
            [
                "(lam_1 (__10 ($0_0)))",
                "(__10 (2))",
            ],
        )



from typing import Callable, Iterator, List, Tuple

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def collect_preorder_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    replace_node_midstream: Callable[
        [SExpression, PreorderMask, int, List[int]], SExpression
    ] = None,
    symbol_to_index_fn=None,
) -> Iterator[Tuple[SExpression, List[str], PreorderMask]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution,
    and yields the S-Expression, the alternate symbols, and the mask at each node.

    :param s_exp: The S-Expression to collect the alternate symbols for.
    :param tree_dist: The tree distribution.
    :param replace_node_midstream: A function that takes the current node, the mask, the
        position of the node, and the alternate symbols that could have been selected, and
        returns the new node. This is used to replace the node as it is being visited. The
        returned node is then visited in place of the original node.
    :param symbol_to_index_fn: A function that takes the mask and the symbol and returns
        the index of the symbol in the tree distribution.
    """
    mask = tree_dist.mask_constructor(tree_dist)
    mask.on_entry(0, 0)
    yield from _collect_preorder_symbols_dfs(
        s_exp,
        tree_dist,
        mask,
        ((0, 0),),
        replace_node_midstream=replace_node_midstream,
        symbol_to_index_fn=symbol_to_index_fn,
    )


def _collect_preorder_symbols_dfs(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    mask: PreorderMask,
    parents: Tuple[Tuple[int, int], ...],
    replace_node_midstream: Callable[
        [SExpression, PreorderMask, int, List[int]], SExpression
    ] = None,
    symbol_to_index_fn=None,
) -> Iterator[Tuple[SExpression, List[str], PreorderMask]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution.
    """
    position = parents[-1][1]
    idxs = np.array([i for i, _ in tree_dist.distribution[parents]])
    bool_mask = mask.compute_mask(position, idxs)
    alts = tuple(int(x) for x in idxs[bool_mask])
    if replace_node_midstream is not None:
        s_exp = replace_node_midstream(s_exp, mask, position, alts)
    yield s_exp, alts, mask
    sym_idx = (
        tree_dist.symbol_to_index[s_exp.symbol]
        if symbol_to_index_fn is None
        else symbol_to_index_fn(mask, s_exp.symbol)
    )
    mask.on_entry(position, sym_idx)
    order = tree_dist.ordering.order(sym_idx, len(s_exp.children))
    for idx, child in zip(order, [s_exp.children[i] for i in order]):
        new_parents = (parents + ((sym_idx, idx),))[-tree_dist.limit :]
        yield from _collect_preorder_symbols_dfs(
            child,
            tree_dist,
            mask,
            new_parents,
            replace_node_midstream=replace_node_midstream,
            symbol_to_index_fn=symbol_to_index_fn,
        )
    mask.on_exit(position, sym_idx)


def annotate_with_alternate_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    summary_fn=lambda chosen, alts: f"{chosen}/{','.join(sorted(alts))}",
) -> SExpression:
    """
    Annotates the S-Expression with the alternate symbols that could have been
    selected in the tree distribution.

    Exists mostly for testing purposes.

    :param s_exp: The S-Expression to annotate.
    :param tree_dist: The tree distribution.
    :param summary_fn: A function that takes the chosen symbol and the alternate
        symbols and returns the new symbol.

    :returns: The annotated S-Expression.
    """
    preorder_symbols = [
        (node, alts) for node, alts, _ in collect_preorder_symbols(s_exp, tree_dist)
    ]
    assert len(preorder_symbols) == len({id(node) for node, _ in preorder_symbols})
    node_id_to_alts = {
        id(node): tuple(tree_dist.symbols[alt][0] for alt in alts)
        for node, alts in preorder_symbols
    }

    def replace(s):
        symbol = summary_fn(s.symbol, node_id_to_alts[id(s)])
        children = [replace(c) for c in s.children]
        return SExpression(symbol, children)

    return replace(s_exp)



import unittest
from fractions import Fraction

import numpy as np
import torch
from parameterized import parameterized

import neurosym as ns
from tests.utils import assertDSL

from .utils import (
    ChildrenInOrderAsserterMask,
    ChildrenInOrderMask,
    ProbabilityTester,
    enumerate_dsl,
)


def get_dsl(with_vars=False, with_3=False):
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    if with_3:
        dslf.concrete("3", "() -> i", lambda: 3)
    if with_vars:
        dslf.concrete("call", "(i -> i, i) -> i", lambda f, x: f(x))
        dslf.lambdas()
    dslf.prune_to("i")
    return dslf.finalize()


def get_dsl_for_ordering():
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i, i) -> t", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("2", "() -> i", lambda: 2)
    dslf.concrete("3", "() -> i", lambda: 2)
    dslf.prune_to("t")
    return dslf.finalize()


def get_more_complex_dsl():
    dslf = ns.DSLFactory()
    dslf.concrete("0", "() -> i", lambda: 0)
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("*", "(f, f) -> f", lambda x, y: x + y)
    dslf.concrete("sqrt", "i -> f", np.sqrt)
    dslf.concrete(">", "(f, f) -> b", np.sqrt)
    dslf.concrete("<", "(i, i) -> b", np.sqrt)
    dslf.concrete("ite", "(b, f, f) -> f", lambda cond, x, y: x if cond else y)
    dslf.lambdas()
    dslf.prune_to("f -> f")
    return dslf.finalize()


dsl = get_dsl()
fam = ns.BigramProgramDistributionFamily(dsl)
dsl_with_3 = get_dsl(with_3=True)
fam_with_3 = ns.BigramProgramDistributionFamily(dsl_with_3)
fam_with_3_no_type_mask = ns.BigramProgramDistributionFamily(
    dsl_with_3, include_type_preorder_mask=False
)
dsl_with_vars = get_dsl(with_vars=True)
fam_with_vars = ns.BigramProgramDistributionFamily(dsl_with_vars)
dsl_for_ordering = get_dsl_for_ordering()
fam_with_ordering = ns.BigramProgramDistributionFamily(
    dsl_for_ordering, additional_preorder_masks=[ChildrenInOrderMask]
)
fam_with_ordering_asserted = ns.BigramProgramDistributionFamily(
    dsl_for_ordering, additional_preorder_masks=[ChildrenInOrderAsserterMask]
)
fam_with_ordering_231 = ns.BigramProgramDistributionFamily(
    dsl_for_ordering,
    additional_preorder_masks=[ChildrenInOrderMask],
    # 2 3 1 means go to index 2 first, then 0 second, then 1 third
    node_ordering=lambda dist: ns.DictionaryNodeOrdering(dist, {"+": [2, 0, 1]}),
)

dsl_more_complex = get_more_complex_dsl()
fam_more_complex = ns.BigramProgramDistributionFamily(dsl_more_complex)

uniform = [
    [
        # root -> each
        [0, 1 / 3, 1 / 3, 1 / 3],
        [0] * 4,
    ],
    [
        # + -> each first arg
        [0, 1 / 3, 1 / 3, 1 / 3],
        # + -> each second arg
        [0, 1 / 3, 1 / 3, 1 / 3],
    ],
    [[0] * 4] * 2,
    [[0] * 4] * 2,
]


class BigramSamplerTest(ProbabilityTester):
    def test_uniform_sample_counts(self):
        n = 10_000
        pdelta = 0.015
        dist = fam.uniform()
        self.assertSameProbDist(dist, uniform)
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 3, pdelta, samples.count("(1)"))
        for plus_two_things in (
            "(+ (1) (1))",
            "(+ (1) (2))",
            "(+ (2) (1))",
            "(+ (2) (2))",
        ):
            self.assertBinomial(
                n, 1 / 3 * 1 / 3 * 1 / 3, pdelta, samples.count(plus_two_things)
            )


class BigramParameterShapeTest(unittest.TestCase):
    def test_shape(self):
        self.assertEqual(fam.parameters_shape(), (4, 2, 4))


class BigramWithParametersGetParametersTest(ProbabilityTester):
    def test_produces_uniform_with_fixed_input(self):
        parameters = torch.zeros((4, 2, 4))
        for off in -1, 298, 29, 3:
            self.assertSameProbDist(
                fam.with_parameters((parameters + off)[None])[0], uniform
            )

    def test_irrelevant_entries_dont_matter(self):
        parameters = torch.zeros((4, 2, 4))
        parameters[0, 0, 0] = 1038
        print(parameters)
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[0, 1, 1] = -3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

        parameters = torch.zeros((4, 2, 4))
        parameters[2] = 3
        self.assertSameProbDist(fam.with_parameters(parameters[None])[0], uniform)

    def test_can_make_deterministic(self):
        parameters = torch.zeros((4, 2, 4))
        # + -> always 1 then 2
        parameters[1, 0, 2] = 100
        parameters[1, 1, 3] = 100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)", "(+ (1) (2))"})

        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(+ (1) (2))"))

    def test_can_make_impossible(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 2, 0.015, samples.count("(2)"))

    def test_can_weight_precisely(self):
        parameters = torch.zeros((4, 2, 4))
        # disable root -> +
        parameters[0, 0, 1] = -100
        # make root -> 1 twice as likely as root -> 2
        parameters[0, 0, 2] = np.log(2)

        dist = fam.with_parameters(parameters[None])[0]
        n = 10_000
        samples = [fam.sample(dist, np.random.RandomState(i)) for i in range(n)]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(1)", "(2)"})

        self.assertBinomial(n, 2 / 3, 0.015, samples.count("(1)"))
        self.assertBinomial(n, 1 / 3, 0.015, samples.count("(2)"))

    def test_fam_with_variables_dsl(self):
        assertDSL(
            self,
            dsl_with_vars.render(),
            """
            $0_0 :: V<i@0>
            $1_0 :: V<i@1>
            $2_0 :: V<i@2>
            $3_0 :: V<i@3>
            + :: (i, i) -> i
            1 :: () -> i
            2 :: () -> i
            call :: (i -> i, i) -> i
            lam :: L<#body|i> -> i -> #body
            """,
        )

    def test_sample_with_variables(self):
        dist = fam_with_vars.uniform()
        n = 20000
        samples = [
            fam_with_vars.sample(dist, np.random.RandomState(i)) for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(1)"))
        self.assertBinomial(n, 1 / 4, 0.01, samples.count("(2)"))
        # see test_call_with_variables for the math
        self.assertBinomial(n, 1 / 80, 0.01, samples.count("(call (lam ($0_0)) (1))"))

    def test_sample_with_ordering(self):
        dist = fam_with_ordering.uniform()
        n = 10_000
        samples = [
            fam_with_ordering.sample(dist, np.random.RandomState(i)) for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(+ (1) (2) (3))"})

    def test_sample_with_ordering_231(self):
        dist = fam_with_ordering_231.uniform()
        n = 10_000
        samples = [
            fam_with_ordering_231.sample(dist, np.random.RandomState(i))
            for i in range(n)
        ]
        samples = [ns.render_s_expression(x) for x in samples]
        self.assertEqual(set(samples), {"(+ (2) (3) (1))"})


class BigramCountProgramsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def convert(self, family, count):
        symbol_text = lambda x: family.tree_distribution_skeleton.symbols[x][0]
        numerators = {
            tuple((symbol_text(sym), pos) for sym, pos in chain): {
                symbol_text(sym): count for sym, count in counts.items()
            }
            for chain, counts in count.numerators.items()
        }
        denominators = {
            tuple((symbol_text(sym), pos) for sym, pos in chain): {
                tuple(symbol_text(sym) for sym in syms): count
                for syms, count in counts.items()
            }
            for chain, counts in count.denominators.items()
        }
        return numerators, denominators

    def count_programs(self, family, programs):
        counts = family.count_programs(
            [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        )
        return [self.convert(family, count) for count in counts.counts]

    def test_counts_single_program(self):
        counts = self.count_programs(fam, [["(+ (1) (2))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                    },
                    {
                        # all symbols could be anything else
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                )
            ],
        )

    def test_counts_multiple_programs(self):
        counts = self.count_programs(fam, [["(+ (1) (2))", "(+ (1) (1))"]])
        print(counts)

        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 2},
                        (("+", 0),): {"1": 2},
                        (("+", 1),): {"2": 1, "1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 2},
                        (("+", 0),): {("+", "1", "2"): 2},
                        (("+", 1),): {("+", "1", "2"): 2},
                    },
                )
            ],
        )

    def test_counts_separate_programs(self):
        counts = self.count_programs(fam, [["(+ (1) (2))"], ["(+ (1) (1))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                ),
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2"): 1},
                        (("+", 0),): {("+", "1", "2"): 1},
                        (("+", 1),): {("+", "1", "2"): 1},
                    },
                ),
            ],
        )

    def test_counts_variables(self):
        counts = self.count_programs(fam_with_vars, [["(call (lam ($0_0)) (1))"]])
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"call": 1},
                        (("call", 0),): {"lam": 1},
                        (("lam", 0),): {"$0_0": 1},
                        (("call", 1),): {"1": 1},
                    },
                    {
                        (("<root>", 0),): {("+", "1", "2", "call"): 1},
                        (("call", 0),): {("lam",): 1},
                        (("lam", 0),): {("$0_0", "+", "1", "2", "call"): 1},
                        (("call", 1),): {("+", "1", "2", "call"): 1},
                    },
                )
            ],
        )

    def test_counts_single_program_ordered(self):
        counts = self.count_programs(fam_with_ordering, [["(+ (1) (2) (3))"]])
        print(counts)
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"1": 1},
                        (("+", 1),): {"2": 1},
                        (("+", 2),): {"3": 1},
                    },
                    {
                        # note that there's no multiplicity here
                        # because the ordering mask doesn't allow it
                        (("<root>", 0),): {("+",): 1},
                        (("+", 0),): {("1",): 1},
                        (("+", 1),): {("2",): 1},
                        (("+", 2),): {("3",): 1},
                    },
                )
            ],
        )

    def test_counts_single_program_ordered_231(self):
        counts = self.count_programs(fam_with_ordering_231, [["(+ (2) (3) (1))"]])
        print(counts)
        self.assertEqual(
            counts,
            [
                (
                    {
                        (("<root>", 0),): {"+": 1},
                        (("+", 0),): {"2": 1},
                        (("+", 1),): {"3": 1},
                        (("+", 2),): {"1": 1},
                    },
                    {
                        # note that there's no multiplicity here
                        # because the ordering mask doesn't allow it
                        (("<root>", 0),): {("+",): 1},
                        (("+", 0),): {("2",): 1},
                        (("+", 1),): {("3",): 1},
                        (("+", 2),): {("1",): 1},
                    },
                )
            ],
        )


class BigramParameterDifferenceLossTest(unittest.TestCase):
    def computeLoss(self, logits, programs, family=fam):
        programs = [[ns.parse_s_expression(x) for x in ps] for ps in programs]
        print(programs)
        loss = family.parameter_difference_loss(
            logits,
            family.count_programs(programs),
        )
        loss = loss.detach().cpu().numpy()
        return loss

    def assertLoss(self, logits, programs, target, **kwargs):
        np.testing.assert_almost_equal(
            self.computeLoss(logits, programs, **kwargs), target
        )

    def assertUniformLoss(self, programs, target, **kwargs):
        logits = torch.zeros((1, 4, 2, 4))
        self.assertLoss(logits, programs, target, **kwargs)
        # fill in impossible location +, 0 -> <root>
        logits[0, 1, 0, 0] = 20
        self.assertLoss(logits, programs, target, **kwargs)

    def test_leaf_program_logit(self):
        self.assertUniformLoss([["(1)"]], [np.log(3)])
        self.assertUniformLoss([["(2)"]], [np.log(3)])

    def test_multiple_program_logit(self):
        self.assertUniformLoss([["(1)", "(2)"]], [np.log(3) * 2])
        self.assertUniformLoss([["(1)", "(1)"]], [np.log(3) * 2])

    def test_separate_program_logit(self):
        self.assertUniformLoss([["(1)"], ["(2)"]], [np.log(3)] * 2)
        self.assertUniformLoss([["(1)"], ["(1)"]], [np.log(3)] * 2)

    def test_nested_program_logit(self):
        self.assertUniformLoss([["(+ (1) (2))"]], [np.log(3) * 3])

    def test_nonuniform_loss(self):
        logits = torch.zeros((2, 4, 2, 4))
        logits[0, 0, 0, 1] = np.log(3)  # 1/2 chance of +
        logits[0, 0, 0, 2] = np.log(2)  # 1/3 chance of 1, 1/6 chance of 2
        self.assertLoss(logits, [["(1)"], ["(1)"]], [-np.log(1 / 3), -np.log(1 / 3)])
        self.assertLoss(logits, [["(2)"], ["(2)"]], [-np.log(1 / 6), -np.log(1 / 3)])
        self.assertLoss(
            logits,
            [["(+ (1) (2))"], ["(+ (1) (2))"]],
            [-(np.log(1 / 2) + np.log(1 / 3) * 2), -3 * np.log(1 / 3)],
        )

    def test_variables_loss(self):
        logits = torch.zeros((1, 10, 2, 10))
        self.assertLoss(logits, [["(1)"]], [np.log(4)], family=fam_with_vars)
        self.assertLoss(
            logits,
            [["(call (lam ($0_0)) (1))"]],
            [np.log(4 * 5 * 4)],
            family=fam_with_vars,
        )


class BigramLikelihoodTest(unittest.TestCase):
    def assertLikelihood(self, dist, program, str_prob, family=fam, **kwargs):
        likelihood = family.compute_likelihood(
            dist, ns.parse_s_expression(program), **kwargs
        )
        self.assertEqual(self.render_likelihood(likelihood), str_prob)

    def assertLikelihoods(self, dist, program, nodes_and_probs, family=fam):
        likelihoods = family.compute_likelihood_per_node(
            dist, ns.parse_s_expression(program)
        )
        likelihoods = [
            (ns.render_s_expression(node), self.render_likelihood(prob))
            for node, prob in likelihoods
        ]
        print(likelihoods)
        self.assertEqual(likelihoods, nodes_and_probs)

    def render_likelihood(self, likelihood):
        prob = np.exp(likelihood)
        prob = Fraction.from_float(float(prob)).limit_denominator(1000)
        result = f"log({prob})"
        return result

    def test_leaf(self):
        self.assertLikelihood(fam.uniform(), "(1)", "log(1/3)")

    def test_plus(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (2))", "log(1/27)")

    def test_plus_nested(self):
        self.assertLikelihood(fam.uniform(), "(+ (1) (+ (1) (2)))", "log(1/243)")
        self.assertLikelihoods(
            fam.uniform(),
            "(+ (1) (+ (1) (2)))",
            [
                ("(+ (1) (+ (1) (2)))", "log(1/3)"),
                ("(1)", "log(1/3)"),
                ("(+ (1) (2))", "log(1/3)"),
                ("(1)", "log(1/3)"),
                ("(2)", "log(1/3)"),
            ],
        )

    def test_leaf_with_variables(self):
        self.assertLikelihood(
            fam_with_vars.uniform(), "(1)", "log(1/4)", family=fam_with_vars
        )

    def test_call_with_variables(self):
        # 1/4 for call
        # 1 for lam
        # 1/5 for $0_0
        # 1/4 for 1
        self.assertLikelihood(
            fam_with_vars.uniform(),
            "(call (lam ($0_0)) (1))",
            "log(1/80)",
            family=fam_with_vars,
        )
        self.assertLikelihoods(
            fam_with_vars.uniform(),
            "(call (lam ($0_0)) (1))",
            [
                ("(call (lam ($0_0)) (1))", "log(1/4)"),
                ("(lam ($0_0))", "log(1)"),
                ("($0_0)", "log(1/5)"),
                ("(1)", "log(1/4)"),
            ],
            family=fam_with_vars,
        )

    def test_ordered(self):
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (1) (2) (3))",
            "log(1)",
            family=fam_with_ordering,
        )
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (2) (3) (1))",
            "log(0)",
            family=fam_with_ordering,
        )

    def test_ordered_tracker(self):
        def tracker(node, likelihood):
            tracked.append(
                (ns.render_s_expression(node), self.render_likelihood(likelihood))
            )

        tracked = []
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (1) (2) (3))",
            "log(1)",
            family=fam_with_ordering,
            tracker=tracker,
        )
        print(tracked)
        self.assertEqual(
            tracked,
            [
                ("(+ (1) (2) (3))", "log(1)"),
                ("(1)", "log(1)"),
                ("(2)", "log(1)"),
                ("(3)", "log(1)"),
            ],
        )

        tracked = []
        self.assertLikelihood(
            fam_with_ordering.uniform(),
            "(+ (2) (3) (1))",
            "log(0)",
            family=fam_with_ordering,
            tracker=tracker,
        )
        print(tracked)
        self.assertEqual(
            tracked,
            [
                ("(+ (2) (3) (1))", "log(1)"),
                ("(2)", "log(0)"),
                ("(3)", "log(0)"),
                ("(1)", "log(0)"),
            ],
        )

    def test_ordered_asserting(self):
        self.assertLikelihood(
            fam_with_ordering_asserted.uniform(),
            "(+ (1) (2) (3))",
            "log(1/27)",
            family=fam_with_ordering_asserted,
        )
        # distribution where 1 is not allowed
        dist = fam_with_ordering_asserted.uniform()
        dist.distribution[:, :, 2] = 0
        self.assertLikelihood(
            dist,
            "(+ (1) (2) (3))",
            "log(0)",
            family=fam_with_ordering_asserted,
        )

    def test_ordered_231(self):
        self.assertLikelihood(
            fam_with_ordering_231.uniform(),
            "(+ (2) (3) (1))",
            "log(1)",
            family=fam_with_ordering_231,
        )
        self.assertLikelihood(
            fam_with_ordering_231.uniform(),
            "(+ (1) (2) (3))",
            "log(0)",
            family=fam_with_ordering_231,
        )

    def test_likelihood_clamped(self, kwargs=lambda x: {}):
        dist = fam.counts_to_distribution(
            fam.count_programs([[ns.parse_s_expression("(1)")]]),
        )[0]
        self.assertEqual(
            fam.compute_likelihood(dist, ns.parse_s_expression("(2)")), -np.inf
        )
        dist = dist.bound_minimum_likelihood(0.01, **kwargs(dist))
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            fam.compute_likelihood(dist, ns.parse_s_expression("(2)")),
            np.log(1 / 100),
            1,
        )
        # should be *very* approximately 1/100 * 1/3 * 1/3
        # even the (+ (1) (1)) should be since the + -> 1 bigram wasn't seen
        # in the original
        # the 1/3 comes from the fact that in the original the + -> 1 and + -> 2 bigrams
        # were undefined, so now they're uniform
        for code in "(+ (1) (2))", "(+ (1) (1))", "(+ (2) (2))":
            self.assertAlmostEqual(
                fam.compute_likelihood(dist, ns.parse_s_expression(code)),
                np.log(1 / 100 * 1 / 3 * 1 / 3),
                1,
            )

    def test_likelihood_clamped_mask_full(self):
        self.test_likelihood_clamped(
            lambda dist: dict(
                symbol_mask=np.ones((dist.distribution.shape[0]), dtype=np.bool_)
            )
        )

    @parameterized.expand([(True,), (False,)])
    def test_likelihood_clamped_mask_partial(self, include_type_mask):
        family = fam_with_3 if include_type_mask else fam_with_3_no_type_mask
        dist = family.counts_to_distribution(
            family.count_programs([[ns.parse_s_expression("(1)")]]),
        )[0]
        self.assertEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(2)")), -np.inf
        )
        self.assertEqual(
            [x for x, _ in family.tree_distribution_skeleton.symbols],
            ["<root>", "+", "1", "2", "3"],
        )
        print(np.array(np.where(dist.distribution)).T)
        dist = dist.bound_minimum_likelihood(
            0.01, symbol_mask=np.array([True, True, True, True, False])
        )
        print(np.array(np.where(dist.distribution)).T)
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(2)")),
            np.log(1 / 100),
            1,
        )
        # should be *very* approximately 1/100
        self.assertAlmostEqual(
            family.compute_likelihood(dist, ns.parse_s_expression("(3)")),
            -np.inf,
            1,
        )

        # see test_likelihood_clamped for the math
        for code in "(+ (1) (2))", "(+ (1) (1))", "(+ (2) (2))":
            self.assertAlmostEqual(
                family.compute_likelihood(dist, ns.parse_s_expression(code)),
                np.log(1 / 100 * 1 / 3 * 1 / 3),
                1,
            )

        # remains -inf because 3 is impossible
        for code in "(+ (1) (3))", "(+ (3) (3))", "(+ (3) (2))":
            self.assertAlmostEqual(
                family.compute_likelihood(dist, ns.parse_s_expression(code)),
                -np.inf,
                1,
            )


class BigramEnumerationTest(unittest.TestCase):
    def test_enumeration_basic(self):
        dist = fam.uniform()

        self.assertEqual(
            enumerate_dsl(fam, dist, -5),
            {
                ("(1)", Fraction(1, 3)),
                ("(2)", Fraction(1, 3)),
                ("(+ (1) (1))", Fraction(1, 27)),
                ("(+ (1) (2))", Fraction(1, 27)),
                ("(+ (2) (1))", Fraction(1, 27)),
                ("(+ (2) (2))", Fraction(1, 27)),
            },
        )

    def test_enumeration_with_vars(self):
        dist = fam_with_vars.uniform()
        self.assertEqual(
            enumerate_dsl(fam_with_vars, dist, -5),
            {
                ("(1)", Fraction(1, 4)),
                ("(2)", Fraction(1, 4)),
                ("(+ (1) (1))", Fraction(1, 64)),
                ("(+ (1) (2))", Fraction(1, 64)),
                ("(+ (2) (1))", Fraction(1, 64)),
                ("(+ (2) (2))", Fraction(1, 64)),
                ("(call (lam ($0_0)) (1))", Fraction(1, 80)),
                ("(call (lam ($0_0)) (2))", Fraction(1, 80)),
                ("(call (lam (1)) (1))", Fraction(1, 80)),
                ("(call (lam (1)) (2))", Fraction(1, 80)),
                ("(call (lam (2)) (1))", Fraction(1, 80)),
                ("(call (lam (2)) (2))", Fraction(1, 80)),
            },
        )

    def test_enumeration_more_complex(self):
        dist = fam_more_complex.uniform()
        self.assertEqual(
            enumerate_dsl(fam_more_complex, dist, -9),
            {
                ("(lam ($0_0))", Fraction(1, 4)),
                ("(lam (sqrt (0)))", Fraction(1, 8)),
                ("(lam (sqrt (+ (0) (0))))", Fraction(1, 32)),
                ("(lam (* ($0_0) ($0_0)))", Fraction(1, 64)),
                ("(lam (* ($0_0) (sqrt (0))))", Fraction(1, 128)),
                ("(lam (* (sqrt (0)) ($0_0)))", Fraction(1, 128)),
                ("(lam (sqrt (+ (+ (0) (0)) (0))))", Fraction(1, 128)),
                ("(lam (sqrt (+ (0) (+ (0) (0)))))", Fraction(1, 128)),
                ("(lam (* (sqrt (0)) (sqrt (0))))", Fraction(1, 256)),
                ("(lam (* ($0_0) (sqrt (+ (0) (0)))))", Fraction(1, 512)),
                ("(lam (* (sqrt (+ (0) (0))) ($0_0)))", Fraction(1, 512)),
                ("(lam (ite (< (0) (0)) ($0_0) ($0_0)))", Fraction(1, 512)),
                ("(lam (sqrt (+ (+ (+ (0) (0)) (0)) (0))))", Fraction(1, 512)),
                ("(lam (sqrt (+ (+ (0) (+ (0) (0))) (0))))", Fraction(1, 512)),
                ("(lam (sqrt (+ (+ (0) (0)) (+ (0) (0)))))", Fraction(1, 512)),
                ("(lam (sqrt (+ (0) (+ (+ (0) (0)) (0)))))", Fraction(1, 512)),
                ("(lam (sqrt (+ (0) (+ (0) (+ (0) (0))))))", Fraction(1, 512)),
                ("(lam (* ($0_0) (* ($0_0) ($0_0))))", Fraction(1, 1024)),
                ("(lam (* (* ($0_0) ($0_0)) ($0_0)))", Fraction(1, 1024)),
                ("(lam (* (sqrt (+ (0) (0))) (sqrt (0))))", Fraction(1, 1024)),
                ("(lam (* (sqrt (0)) (sqrt (+ (0) (0)))))", Fraction(1, 1024)),
                ("(lam (ite (< (0) (0)) ($0_0) (sqrt (0))))", Fraction(1, 1024)),
                ("(lam (ite (< (0) (0)) (sqrt (0)) ($0_0)))", Fraction(1, 1024)),
                ("(lam (* ($0_0) (* ($0_0) (sqrt (0)))))", Fraction(1, 2048)),
                ("(lam (* ($0_0) (* (sqrt (0)) ($0_0))))", Fraction(1, 2048)),
                ("(lam (* ($0_0) (sqrt (+ (+ (0) (0)) (0)))))", Fraction(1, 2048)),
                ("(lam (* ($0_0) (sqrt (+ (0) (+ (0) (0))))))", Fraction(1, 2048)),
                ("(lam (* (* ($0_0) ($0_0)) (sqrt (0))))", Fraction(1, 2048)),
                ("(lam (* (* ($0_0) (sqrt (0))) ($0_0)))", Fraction(1, 2048)),
                ("(lam (* (* (sqrt (0)) ($0_0)) ($0_0)))", Fraction(1, 2048)),
                ("(lam (* (sqrt (+ (+ (0) (0)) (0))) ($0_0)))", Fraction(1, 2048)),
                ("(lam (* (sqrt (+ (0) (+ (0) (0)))) ($0_0)))", Fraction(1, 2048)),
                ("(lam (* (sqrt (0)) (* ($0_0) ($0_0))))", Fraction(1, 2048)),
                ("(lam (ite (< (+ (0) (0)) (0)) ($0_0) ($0_0)))", Fraction(1, 2048)),
                ("(lam (ite (< (0) (+ (0) (0))) ($0_0) ($0_0)))", Fraction(1, 2048)),
                ("(lam (ite (< (0) (0)) (sqrt (0)) (sqrt (0))))", Fraction(1, 2048)),
                ("(lam (ite (> ($0_0) ($0_0)) ($0_0) ($0_0)))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (+ (+ (0) (0)) (0)) (0)) (0))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (+ (0) (+ (0) (0))) (0)) (0))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (+ (0) (0)) (+ (0) (0))) (0))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (+ (0) (0)) (0)) (+ (0) (0)))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (0) (+ (+ (0) (0)) (0))) (0))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (0) (+ (0) (+ (0) (0)))) (0))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (0) (+ (0) (0))) (+ (0) (0)))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (0) (0)) (+ (+ (0) (0)) (0)))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (+ (0) (0)) (+ (0) (+ (0) (0))))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (0) (+ (+ (+ (0) (0)) (0)) (0)))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (0) (+ (+ (0) (+ (0) (0))) (0)))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (0) (+ (+ (0) (0)) (+ (0) (0))))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (0) (+ (0) (+ (+ (0) (0)) (0))))))", Fraction(1, 2048)),
                ("(lam (sqrt (+ (0) (+ (0) (+ (0) (+ (0) (0)))))))", Fraction(1, 2048)),
                ("(lam (* ($0_0) (* (sqrt (0)) (sqrt (0)))))", Fraction(1, 4096)),
                ("(lam (* (* ($0_0) (sqrt (0))) (sqrt (0))))", Fraction(1, 4096)),
                ("(lam (* (* (sqrt (0)) ($0_0)) (sqrt (0))))", Fraction(1, 4096)),
                ("(lam (* (* (sqrt (0)) (sqrt (0))) ($0_0)))", Fraction(1, 4096)),
                ("(lam (* (sqrt (+ (+ (0) (0)) (0))) (sqrt (0))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (+ (0) (+ (0) (0)))) (sqrt (0))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (+ (0) (0))) (sqrt (+ (0) (0)))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (0)) (* ($0_0) (sqrt (0)))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (0)) (* (sqrt (0)) ($0_0))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (0)) (sqrt (+ (+ (0) (0)) (0)))))", Fraction(1, 4096)),
                ("(lam (* (sqrt (0)) (sqrt (+ (0) (+ (0) (0))))))", Fraction(1, 4096)),
                (
                    "(lam (ite (< (+ (0) (0)) (0)) ($0_0) (sqrt (0))))",
                    Fraction(1, 4096),
                ),
                (
                    "(lam (ite (< (+ (0) (0)) (0)) (sqrt (0)) ($0_0)))",
                    Fraction(1, 4096),
                ),
                (
                    "(lam (ite (< (0) (+ (0) (0))) ($0_0) (sqrt (0))))",
                    Fraction(1, 4096),
                ),
                (
                    "(lam (ite (< (0) (+ (0) (0))) (sqrt (0)) ($0_0)))",
                    Fraction(1, 4096),
                ),
                (
                    "(lam (ite (< (0) (0)) ($0_0) (sqrt (+ (0) (0)))))",
                    Fraction(1, 4096),
                ),
                (
                    "(lam (ite (< (0) (0)) (sqrt (+ (0) (0))) ($0_0)))",
                    Fraction(1, 4096),
                ),
                ("(lam (ite (> ($0_0) ($0_0)) ($0_0) (sqrt (0))))", Fraction(1, 4096)),
                ("(lam (ite (> ($0_0) ($0_0)) (sqrt (0)) ($0_0)))", Fraction(1, 4096)),
                ("(lam (ite (> ($0_0) (sqrt (0))) ($0_0) ($0_0)))", Fraction(1, 4096)),
                ("(lam (ite (> (sqrt (0)) ($0_0)) ($0_0) ($0_0)))", Fraction(1, 4096)),
            },
        )

    def test_enumeration_more_complex_get_many(self):
        dist = fam_more_complex.uniform()
        large_set = sorted(
            enumerate_dsl(fam_more_complex, dist, -12), key=lambda x: (-x[1], x[0])
        )
        self.assertEqual(len(large_set), 1609)
        print(large_set[::100])
        self.assertEqual(
            large_set[::100],
            [
                ("(lam ($0_0))", Fraction(1, 4)),
                ("(lam (ite (< (0) (0)) (* ($0_0) ($0_0)) ($0_0)))", Fraction(1, 8192)),
                (
                    "(lam (ite (< (0) (+ (0) (0))) ($0_0) (sqrt (+ (0) (0)))))",
                    Fraction(1, 16384),
                ),
                (
                    "(lam (ite (< (+ (+ (0) (0)) (+ (0) (0))) (0)) ($0_0) ($0_0)))",
                    Fraction(1, 32768),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (0) (0)) (0)) (+ (0) (+ (0) (+ (0) (0)))))))",
                    Fraction(1, 32768),
                ),
                (
                    "(lam (* ($0_0) (ite (< (0) (0)) ($0_0) (sqrt (+ (0) (0))))))",
                    Fraction(1, 65536),
                ),
                (
                    "(lam (* (sqrt (0)) (ite (< (0) (+ (0) (0))) ($0_0) ($0_0))))",
                    Fraction(1, 65536),
                ),
                (
                    "(lam (ite (> ($0_0) ($0_0)) (* ($0_0) (sqrt (0))) ($0_0)))",
                    Fraction(1, 65536),
                ),
                (
                    "(lam (* ($0_0) (sqrt (+ (0) (+ (+ (+ (+ (0) (0)) (0)) (0)) (0))))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (* (sqrt (+ (+ (+ (0) (+ (+ (0) (0)) (0))) (0)) (0))) ($0_0)))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (ite (< (+ (0) (+ (0) (+ (0) (0)))) (+ (0) (0))) ($0_0) ($0_0)))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (ite (< (0) (0)) (sqrt (0)) (ite (< (0) (0)) ($0_0) ($0_0))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (+ (+ (0) (0)) (0)) (+ (0) (0))) (+ (0) (0))) (0))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (0) (+ (0) (0))) (0)) (+ (+ (0) (0)) (+ (0) (0))))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (sqrt (+ (+ (0) (+ (0) (+ (+ (0) (+ (0) (0))) (+ (0) (0))))) (0))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (sqrt (+ (0) (+ (+ (+ (0) (+ (+ (0) (0)) (0))) (+ (0) (0))) (0)))))",
                    Fraction(1, 131072),
                ),
                (
                    "(lam (sqrt (+ (0) (+ (0) (+ (0) (+ (+ (0) (+ (0) (+ (0) (0)))) (0)))))))",
                    Fraction(1, 131072),
                ),
            ],
        )

    def test_enumeration_more_complex_get_even_more(self):
        dist = fam_more_complex.uniform()
        large_set = sorted(
            enumerate_dsl(fam_more_complex, dist, -15, max_denominator=10**8),
            key=lambda x: (-x[1], x[0]),
        )
        self.assertEqual(len(large_set), 18804)
        print(large_set[::1000])
        self.assertEqual(
            large_set[::1000],
            [
                ("(lam ($0_0))", Fraction(1, 4)),
                (
                    "(lam (ite (< (+ (0) (+ (0) (+ (0) (0)))) (+ (0) (0))) ($0_0) ($0_0)))",
                    Fraction(707, 92667907),
                ),
                (
                    "(lam (* (sqrt (0)) (sqrt (+ (+ (+ (+ (0) (0)) (0)) (0)) (+ (0) (0))))))",
                    Fraction(334, 87556099),
                ),
                (
                    "(lam (* (ite (> ($0_0) (sqrt (0))) (sqrt (+ (0) (0))) ($0_0)) ($0_0)))",
                    Fraction(158, 82837507),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (+ (+ (+ (0) (+ (0) (+ (0) (0)))) (0)) (0)) (0)) (0)) (0))))",
                    Fraction(158, 82837507),
                ),
                (
                    "(lam (sqrt (+ (0) (+ (+ (+ (+ (+ (0) (+ (0) (+ (0) (0)))) (0)) (0)) (0)) (0)))))",
                    Fraction(158, 82837507),
                ),
                (
                    "(lam (* (* (sqrt (0)) ($0_0)) (sqrt (+ (0) (+ (+ (+ (0) (0)) (0)) (0))))))",
                    Fraction(25, 26214401),
                ),
                (
                    "(lam (ite (< (+ (+ (+ (0) (0)) (0)) (0)) (+ (0) (0))) (sqrt (+ (0) (0))) ($0_0)))",
                    Fraction(25, 26214401),
                ),
                (
                    "(lam (ite (> ($0_0) ($0_0)) (sqrt (+ (0) (+ (0) (+ (0) (+ (0) (0)))))) ($0_0)))",
                    Fraction(25, 26214401),
                ),
                (
                    "(lam (* ($0_0) (sqrt (+ (+ (+ (0) (0)) (0)) (+ (+ (0) (+ (0) (+ (0) (0)))) (0))))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (* (ite (< (+ (0) (+ (0) (0))) (0)) ($0_0) (sqrt (+ (0) (0)))) (sqrt (0))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (* (sqrt (+ (0) (0))) (ite (< (0) (0)) (* ($0_0) (sqrt (0))) ($0_0))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (ite (< (+ (0) (0)) (0)) (* ($0_0) (sqrt (0))) (* ($0_0) (sqrt (0)))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (ite (< (0) (0)) (sqrt (0)) (sqrt (+ (+ (0) (+ (0) (+ (+ (0) (0)) (0)))) (0)))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (+ (+ (+ (0) (+ (0) (+ (0) (0)))) (0)) (0)) (0)) (0)) (+ (0) (0)))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (sqrt (+ (+ (+ (0) (+ (+ (+ (0) (+ (0) (0))) (0)) (0))) (+ (+ (0) (0)) (0))) (0))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (sqrt (+ (+ (0) (+ (+ (+ (+ (0) (+ (0) (0))) (0)) (+ (0) (0))) (0))) (+ (0) (0)))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (sqrt (+ (+ (0) (0)) (+ (+ (+ (+ (0) (0)) (+ (0) (+ (0) (0)))) (0)) (+ (0) (0))))))",
                    Fraction(12, 25165825),
                ),
                (
                    "(lam (sqrt (+ (0) (+ (+ (0) (+ (+ (+ (0) (0)) (+ (0) (+ (0) (0)))) (+ (0) (0)))) (0)))))",
                    Fraction(12, 25165825),
                ),
            ],
        )


class BigramMixTest(unittest.TestCase):
    def fit_family_to_programs(self, family, programs):
        return family.counts_to_distribution(
            family.count_programs([[ns.parse_s_expression(x) for x in programs]]),
        )[0]

    def merged(self, fam_1, programs_1, fam_2, programs_2, weight):
        dist_1 = self.fit_family_to_programs(fam_1, programs_1)
        dist_2 = self.fit_family_to_programs(fam_2, programs_2)
        return dist_1.mix_with_other(dist_2, weight)

    def assertClose(self, dist_merged, dist_expected):
        delta = dist_merged.distribution - dist_expected.distribution
        bad = np.abs(delta) > 1e-6
        if bad.any():
            syms = dist_merged.dist_fam.symbols()
            for parent, which in np.array(np.where(bad.any(-1))).T:
                actual = dist_merged.distribution[parent, which]
                expected = dist_expected.distribution[parent, which]
                select = (actual > 0) | (expected > 0)
                actual, expected = [
                    {syms[i]: arr[i] for i in np.where(select)[0]}
                    for arr in (actual, expected)
                ]
                self.assertEqual(
                    actual, expected, f"At {syms[parent]}'s {which}th child"
                )

    def assertMix(self, fam_1, programs_1, fam_2, programs_2, weight_1, weight_2):
        self.assertClose(
            self.merged(
                fam_1, programs_1, fam_2, programs_2, weight_2 / (weight_1 + weight_2)
            ),
            self.fit_family_to_programs(
                fam_1 if len(fam_1.symbols()) > len(fam_2.symbols()) else fam_2,
                programs_1 * weight_1 + programs_2 * weight_2,
            ),
        )

    def assertMixBoth(self, fam_1, programs_1, fam_2, programs_2, weight_1, weight_2):
        # pylint: disable=arguments-out-of-order
        self.assertMix(fam_1, programs_1, fam_2, programs_2, weight_1, weight_2)
        self.assertMix(fam_2, programs_2, fam_1, programs_1, weight_2, weight_1)

    progs_1 = "(1)", "(+ (1) (2))"
    progs_2 = "(+ (2) (2))", "(2)"
    progs_3 = "(+ (2) (3))", "(3)"

    def test_merge_bigrams_same_dsl_equal_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 1, 1)

    def test_merge_bigrams_same_dsl_more_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 2, 1)

    def test_merge_bigrams_same_dsl_less_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam, self.progs_2, 2, 5)

    def test_merge_bigrams_with_different_fams_equal_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 1, 1)

    def test_merge_bigrams_with_different_fams_more_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 9, 4)

    def test_merge_bigrams_with_different_fams_less_weight(self):
        self.assertMixBoth(fam, self.progs_1, fam_with_3, self.progs_3, 2, 4)

    def test_non_subset_fails(self):
        with self.assertRaises(ValueError) as cm:
            self.merged(fam_with_3, self.progs_3, fam_with_vars, self.progs_1, 1)

        self.assertEqual(
            str(cm.exception),
            "DSL not compatible, extra symbols in this: '3', "
            "extra symbols in other: '$0_0', '$1_0', '$2_0', '$3_0', 'call', 'lam'",
        )


class OrderingTest(unittest.TestCase):
    def test_traversal(self):
        ordering = fam_with_ordering.tree_distribution_skeleton.ordering
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (1) (2) (3))")
                )
            ],
            ["(+ (1) (2) (3))", "(1)", "(2)", "(3)"],
        )

    def test_traversal_231(self):
        ordering = fam_with_ordering_231.tree_distribution_skeleton.ordering
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (2) (3) (1))")
                )
            ],
            ["(+ (2) (3) (1))", "(1)", "(2)", "(3)"],
        )
        self.assertEqual(
            [
                ns.render_s_expression(x)
                for x in ordering.traverse_preorder(
                    ns.parse_s_expression("(+ (1) (2) (3))")
                )
            ],
            ["(+ (1) (2) (3))", "(3)", "(1)", "(2)"],
        )



from dataclasses import dataclass
from typing import Callable

import torch
from sklearn.metrics import f1_score, hamming_loss
from torch import nn

from .base_trainer import BaseTrainer, BaseTrainerConfig


def classification_mse_loss(predictions, targets):
    # pylint: disable=not-callable
    targets = torch.nn.functional.one_hot(
        targets.squeeze(-1), num_classes=predictions.shape[-1]
    ).float()
    predictions = predictions.view(-1, predictions.shape[-1])
    targets = targets.view(-1, targets.shape[-1])
    return nn.functional.mse_loss(predictions, targets)


@dataclass
class NEARTrainerConfig(BaseTrainerConfig):
    """
    Configuration class for training a NEAR model.

    :param lr: Learning rate for the optimizer (default: 1e-4)
    :param weight_decay: Weight decay for the optimizer (default: 0.0)
    :param n_epochs: Number of epochs to train for (default: 10)
    :param seed: Random seed for reproducibility (default: 44)
    :param evaluate: Whether to evaluate the model after training (default: True)
    :param resume: Path to a checkpoint to resume training from (default: "")
    :param scheduler: Learning rate scheduler to use (default: "cosine")
    :param sav_dir: Directory to save checkpoints to (default: "data/shapeworldonly_checkpoints")
    :param optimizer: Optimizer to use (default: torch.optim.Adam)
    :param max_seq_len: Maximum sequence length for the model (default: 100)
    :param loss_callback: Loss function to use (default: ``classification_mse_loss``)
    """

    max_seq_len: int = 100
    loss_callback: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = (
        classification_mse_loss
    )
    num_labels: int = -1  # Set Programmatically


class NEARTrainer(BaseTrainer):
    """
    Trainer class for training a NEAR model. Is a pl.LightningModule.

    :param model: The module to train
    :param config: Configuration for the trainer
    """

    def __init__(self, model: nn.Module, config: NEARTrainerConfig):
        super().__init__(model=model, config=config)
        assert config.num_labels > 0, "Number of labels must be set programmatically"
        self.loss_fn = config.loss_callback

    @classmethod
    def _compute_average_f1_score(
        cls, predictions: torch.Tensor, targets: torch.Tensor, num_labels: int
    ):  # noqa: E501
        if num_labels > 1:
            weighted_avg_f1 = 1 - f1_score(targets, predictions, average="weighted")
            unweighted_avg_f1 = 1 - f1_score(targets, predictions, average="macro")
            all_f1 = 1 - f1_score(targets, predictions, average=None)
            return dict(
                weighted_avg_f1=weighted_avg_f1,
                unweighted_avg_f1=unweighted_avg_f1,
                all_f1s=all_f1,
            )
        avg_f1 = 1 - f1_score(targets, predictions, average="binary")
        all_f1 = 1 - f1_score(targets, predictions, average=None)
        return dict(avg_f1=avg_f1, all_f1s=all_f1)

    @classmethod
    def label_correctness(
        cls, predictions: torch.Tensor, targets: torch.Tensor, num_labels: int
    ):  # noqa: E501
        hamming_accuracy = 1 - hamming_loss(
            targets.squeeze().cpu(), predictions.squeeze().cpu()
        )
        f1_scores = cls._compute_average_f1_score(
            predictions, targets, num_labels
        )  # noqa: E501
        return dict(hamming_accuracy=hamming_accuracy, **f1_scores)

    def loss(self, predictions: torch.Tensor, targets: torch.Tensor) -> dict:
        # pylint: disable=arguments-differ
        loss = self.loss_fn(predictions, targets)
        return loss

    def _step(self, inputs, outputs, validation=False, **kwargs):
        # pylint: disable=arguments-differ
        del kwargs
        predictions = self.model(inputs)
        losses = dict(loss=self.loss(predictions, outputs))

        if validation and self.logger is not None and self.current_epoch % 2 == 0:
            self._log_program_accuracy(predictions, outputs, inputs)

        return losses

    def _log_program_accuracy(self, predictions, outputs, inputs):
        del inputs
        if self.config.num_labels > 1:
            predictions = torch.argmax(predictions, dim=-1)
        else:
            predictions = torch.round(torch.sigmoid(predictions))
        targets = outputs
        correctness = NEARTrainer.label_correctness(
            predictions, targets, self.config.num_labels
        )
        self.logger.log_metrics(correctness, step=self.global_step)


def _main():
    from neurosym.utils.imports import import_pytorch_lightning

    pl = import_pytorch_lightning()
    from neurosym.datasets.load_data import DatasetFromNpy, DatasetWrapper

    def dataset_factory(train_seed):
        return DatasetWrapper(
            DatasetFromNpy(
                "../data/classification_example/train_ex_data.npy",
                "../data/classification_example/train_ex_labels.npy",
                train_seed,
            ),
            DatasetFromNpy(
                "../data/classification_example/test_ex_data.npy",
                "../data/classification_example/test_ex_labels.npy",
                None,
            ),
            batch_size=200,
        )

    datamodule = dataset_factory(42)
    # model = nn.Sequential(
    #     nn.Linear(2, 100),
    #     nn.ReLU(),
    #     nn.Linear(100, 2),
    # )
    model = nn.Linear(2, 2)
    # model.weight.data = torch.tensor([[0., 1.], [0., 0.]])
    # model.bias.data = torch.tensor([0., 0.])

    def nll_loss(predictions, targets):
        predictions = (
            predictions.view(-1, predictions.shape[-1])
            .clamp(min=1e-10)
            .log()
            .log_softmax(dim=-1)
        )
        targets = targets.view(-1)
        return nn.functional.nll_loss(predictions, targets)

    trainer_cfg = NEARTrainerConfig(
        lr=1e-4,
        max_seq_len=100,
        n_epochs=10000,
        num_labels=2,
        train_steps=len(datamodule.train),
        loss_callback=nll_loss,
    )
    pl_model = NEARTrainer(model, config=trainer_cfg)
    trainer = pl.Trainer(
        max_epochs=trainer_cfg.n_epochs,
        devices="auto",
        accelerator="cpu",
        enable_checkpointing=False,
        enable_model_summary=False,
        logger=False,
        callbacks=[],
    )
    trainer.fit(pl_model, datamodule.train_dataloader(), datamodule.val_dataloader())
    trainer.validate(pl_model, datamodule.val_dataloader())

    print(trainer.callback_metrics)


if __name__ == "__main__":
    _main()



import json
import unittest

from tests.tutorial.utils import execute_notebook
from tutorial.process_tutorial import create_skeleton


class TestDiscreteExercise(unittest.TestCase):
    def test_discrete_exercise_skeleton(self):
        with open("tutorial/discrete_exercise_solutions.ipynb") as f:
            solutions = json.load(f)
        with open("tutorial/discrete_exercise_skeleton.ipynb") as f:
            notebook = json.load(f)

        self.assertEqual(
            create_skeleton(solutions, "tutorial/discrete_exercise_solutions.ipynb"),
            notebook,
        )

    def test_discrete_exercise_skeleton_runnable(self):
        result = execute_notebook(
            "tutorial/discrete_exercise_solutions.ipynb",
            suffix="import json; print('*' * 80); print(json.dumps(abstraction_dsl.render()))",
        )
        *_, stars, actual, _ = result.split("\n")
        self.assertEqual(stars, "*" * 80)
        actual = json.loads(actual)
        expected = """
              0 :: () -> f
              1 :: () -> f
              2 :: () -> f
              + :: (f, f) -> f
              - :: (f, f) -> f
              * :: (f, f) -> f
             ** :: (f, f) -> f
              / :: (f, f) -> f
            sin :: f -> f
           sqrt :: f -> f
              < :: (f, f) -> b
            ite :: (b, f, f) -> f
            lam :: L<#body|f> -> f -> #body
           $0_0 :: V<f@0>
           __10 :: (f, f) -> f = (lam-abstr (#0 #1) (sin (/ #1 #0)))
           __20 :: () -> f = (lam-abstr () (sqrt (2)))
           __30 :: (f, f) -> f -> f = (lam-abstr (#0 #1) (lam (- #1 (sin (** #0 ($0_0))))))
           __40 :: (f, f) -> f = (lam-abstr (#0 #1) (ite (< #1 #0) (1) (0)))
           __50 :: () -> f = (lam-abstr () (sin (1)))
        """
        actual_lines, expected_lines = [
            [line.strip() for line in text.split("\n") if line.strip()]
            for text in (actual, expected)
        ]
        self.assertEqual(actual_lines, expected_lines)



from abc import ABC, abstractmethod
from typing import List

from ..programs.hole import Hole, _all_holes
from ..programs.s_expression import SExpression


class HoleSetChooser(ABC):
    """
    Utility for the DSL search graph, which tells it how to choose sets of holes to expand.
    """

    @abstractmethod
    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        """
        Returns a list of sets of holes, where each set of holes will be expanded
        simultaneously. The order of the sets of holes is the order in which they
        will be expanded.
        """


class ChooseFirst(HoleSetChooser):
    """
    Hole set chooser that chooses the first hole it finds.
    """

    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        result = []
        for hole in _all_holes(node):
            result.append([hole])
            break
        return result


class ChooseAll(HoleSetChooser):
    """
    Hole set chooser that chooses all holes.
    """

    def choose_hole_sets(self, node: SExpression) -> List[List[Hole]]:
        result = []
        for hole in _all_holes(node):
            result.append([hole])
        return result



import ast
import glob
import json
import os
import re
import unittest
from functools import lru_cache
from types import ModuleType

import parameterized

import neurosym as ns
from neurosym.examples import near
from tests.tutorial.utils import ipynb_to_py


def files_to_examine(*paths):
    for path in paths:
        for root, _, files in os.walk(path):
            for file in files:
                if ".ipynb_checkpoints" in root:
                    continue
                if file.endswith(".py") or file.endswith(".ipynb"):
                    yield os.path.join(root, file)


def read_python_file(path):
    with open(path) as f:
        text = f.read()
    if path.endswith(".py"):
        return text
    if path.endswith(".ipynb"):
        code = ipynb_to_py(json.loads(text))
        code = "\n".join(code)
        return code
    raise ValueError(f"Unknown file type: {path}")


class GatherImports(ast.NodeVisitor):
    def __init__(self):
        self.imports = set()

    def visit_Import(self, node):
        self.imports.add(node)

    def visit_ImportFrom(self, node):
        self.imports.add(node)


class OnlyDirectImportsTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [(path,) for path in files_to_examine("tests", "tutorial")]
    )
    def test_only_direct_import(self, path):
        code = read_python_file(path)
        tree = ast.parse(code)
        gatherer = GatherImports()
        gatherer.visit(tree)
        imports = {ast.unparse(node) for node in gatherer.imports}
        imports = {imp for imp in imports if "neurosym" in imp}

        expected = {"from neurosym.examples import near", "import neurosym as ns"}

        self.assertEqual(imports | expected, expected)


class NoPrintsTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [(path,) for path in files_to_examine("neurosym")]
    )
    def test_no_prints(self, path):
        if path in {"neurosym/examples/near/methods/near_example_trainer.py"}:
            # skip this file, it is an example
            return
        if path in {"neurosym/utils/logging.py"}:
            # skip this file, it is a logging utility
            return
        with open(path) as f:
            code = f.read()
        code = ast.parse(code)
        nodes = ast.walk(code)
        nodes = [node for node in nodes if isinstance(node, ast.Name)]
        nodes = [node for node in nodes if node.id == "print"]
        self.assertEqual(len(nodes), 0)


@lru_cache(None)
def documented_functions():
    rst_files = glob.glob("docs/source/*.rst")
    functions = set()
    for rst_file in rst_files:
        with open(rst_file) as f:
            text = f.read()
        functions.update(
            x.group("name")
            for x in re.finditer(
                r".. auto(function|class|attribute):: (?P<name>.*)", text
            )
        )
    return functions


class AllFunctionsDocumentedTest(unittest.TestCase):

    pattern = re.compile(r"\b(ns|near)(\.(\w+))+")

    def normalize(self, function):
        if function.startswith("ns."):
            function = function.replace("ns.", "neurosym.", 1)
        if function.startswith("near."):
            function = function.replace("near.", "neurosym.examples.near.", 1)
        function = {
            "neurosym.examples.near.NeuralDSL.from_dsl": "neurosym.examples.near.NeuralDSL",
            "neurosym.PythonDSLSubset.from_s_exps": "neurosym.PythonDSLSubset",
            "neurosym.PythonDSLSubset.from_programs": "neurosym.PythonDSLSubset",
            "neurosym.Environment.empty": "neurosym.Environment",
        }.get(function, function)
        return function

    @parameterized.parameterized.expand(
        [(path,) for path in files_to_examine("tests", "tutorial")]
    )
    def test_all_used_functions_documented(self, path):
        code = read_python_file(path)
        functions = set(x.group() for x in self.pattern.finditer(code))
        functions = {self.normalize(f) for f in functions}
        print(documented_functions())
        extras = functions - documented_functions()
        if not extras:
            return
        self.fail(f"Found {len(extras)} undocumented functions: {extras}")

    def assertFieldDocumented(self, name, base_module, base_module_name):
        if name.startswith("_"):
            return
        if isinstance(getattr(base_module, name), ModuleType):
            return
        self.assertIn(base_module_name + "." + name, documented_functions())

    @parameterized.parameterized.expand([(name,) for name in dir(ns)])
    def test_ns_function_documented(self, name):
        self.assertFieldDocumented(name, ns, "neurosym")

    @parameterized.parameterized.expand([(name,) for name in dir(near)])
    def test_near_function_documented(self, name):
        self.assertFieldDocumented(name, near, "neurosym.examples.near")



"""
We want to enumerate programs as trees. Specifically, we have some distribution
    P(tree) = prod_{node in tree} P(node | ancestors(node)[:limit]).

Here, limit is a value that the distribution guarantees us. This is useful
    because we can use this to optimize our enumeration algorithm (since it
    reduces the amount of memory we need to save for each hole).

Our algorithm here is based on iterative deepening. We have a method that
    keeps a minimum and maximum likelihood, and enumerates all programs
    that are within this likelihood range.

Likelihood is defined as the log probability of the program.
"""

import bisect
import itertools
from typing import Any, Dict, List, Tuple, Union

import numpy as np
from torch import NoneType

from neurosym.program_dist.enumeration_chunk_size import DEFAULT_CHUNK_SIZE
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def enumerate_tree_dist(
    tree_dist: TreeDistribution,
    *,
    chunk_size: float = DEFAULT_CHUNK_SIZE,
    min_likelihood: float = float("-inf"),
    use_cache=True,
):
    """
    Enumerate all programs using iterative deepening.

    :param tree_dist: The distribution to sample from.
    :param min_likelihood: The minimum likelihood to consider.
    :param chunk_size: The amount of likelihood to consider at once. If this is
        too small, we will spend a lot of time doing the same work over and
        over again. If this is too large, we will spend a lot of time
        doing work that we don't need to do.
    :param use_cache: Whether to use a cache to store intermediate results.
        This will be disabled if the mask does not support caching.
    """
    for chunk in itertools.count(1):
        likelihood_bound = -chunk * chunk_size
        preorder_mask = tree_dist.mask_constructor(tree_dist)
        cache = {} if use_cache and preorder_mask.can_cache else None
        preorder_mask.on_entry(0, 0)
        for program, likelihood in _enumerate_tree_dist_dfs(
            tree_dist, likelihood_bound, ((0, 0),), preorder_mask, cache
        ):
            if (
                max(likelihood_bound, min_likelihood)
                < likelihood
                <= likelihood_bound + chunk_size
            ):
                yield program, likelihood
        if likelihood_bound <= min_likelihood:
            return


def _remove_below_threshold(
    results: List[Tuple[SExpression, float]], min_likelihood: float
):
    """
    Remove all results below the threshold.
    """
    # binary search
    index = bisect.bisect_left(results, min_likelihood, key=lambda x: x[1])
    return results[index:]


def _enumerate_tree_dist_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
    cache: Union[NoneType, Dict[Any, List[Tuple[SExpression, float]]]],
):
    if cache is not None:
        key = preorder_mask.cache_key(parents), parents
        if key in cache:
            old_results, old_min_likelihood = cache[key]
            if old_min_likelihood <= min_likelihood:
                return _remove_below_threshold(old_results, min_likelihood)
    generator = _enumerate_tree_dist_dfs_uncached(
        tree_dist, min_likelihood, parents, preorder_mask, cache
    )
    if cache is not None:
        generator = sorted(generator, key=lambda x: x[1])
        cache[key] = generator, min_likelihood
    return generator


def _enumerate_tree_dist_dfs_uncached(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
    cache: Union[NoneType, Dict[Any, List[Tuple[SExpression, float]]]],
):
    """
    Enumerate all programs that are within the likelihood range, with the given parents.
    """

    if min_likelihood > 0:
        # We can stop searching deeper.
        return

    assert len(parents) <= tree_dist.limit
    assert isinstance(
        preorder_mask, PreorderMask
    ), f"{preorder_mask} is not a PreorderMask"

    # Performed recursively for now.
    syms, log_probs = tree_dist.likelihood_arrays[parents]
    position = parents[-1][1]
    mask = preorder_mask.compute_mask(position, syms)
    denominator = np.logaddexp.reduce(log_probs[mask])
    for node, likelihood in zip(syms[mask], log_probs[mask] - denominator):
        likelihood = np.float64(likelihood)
        new_parents = parents + (node,)
        new_parents = new_parents[-tree_dist.limit :]
        symbol, arity = tree_dist.symbols[node]
        undo_entry = preorder_mask.on_entry(position, node)
        for children, child_likelihood in _enumerate_children_and_likelihoods_dfs(
            tree_dist,
            min_likelihood - likelihood,
            parents,
            node,
            num_children=arity,
            starting_index=0,
            order=tree_dist.ordering.order(node, arity),
            preorder_mask=preorder_mask,
            cache=cache,
        ):
            if child_likelihood + likelihood < min_likelihood:
                continue
            undo_exit = preorder_mask.on_exit(position, node)
            yield SExpression(
                symbol, [children[i] for i in range(arity)]
            ), child_likelihood + likelihood
            undo_exit()
        undo_entry()


def _enumerate_children_and_likelihoods_dfs(
    tree_dist: TreeDistribution,
    min_likelihood: float,
    parents: Tuple[Tuple[int, int], ...],
    most_recent_parent: int,
    num_children: int,
    starting_index: int,
    order: List[int],
    preorder_mask: PreorderMask,
    cache: Union[NoneType, Dict[Any, List[Tuple[SExpression, float]]]],
):
    """
    Enumerate all children and their likelihoods.
    """

    if min_likelihood > 0:
        # We can stop searching deeper.
        return

    if starting_index == num_children:
        yield {}, 0
        return
    new_parents = parents + ((most_recent_parent, order[starting_index]),)
    new_parents = new_parents[-tree_dist.limit :]

    for first_child, first_likelihood in _enumerate_tree_dist_dfs(
        tree_dist, min_likelihood, new_parents, preorder_mask, cache
    ):
        for (
            rest_children,
            rest_likelihood,
        ) in _enumerate_children_and_likelihoods_dfs(
            tree_dist,
            min_likelihood - first_likelihood,
            parents,
            most_recent_parent,
            num_children,
            starting_index + 1,
            order,
            preorder_mask,
            cache=cache,
        ):
            rest_children[order[starting_index]] = first_child
            yield rest_children, first_likelihood + rest_likelihood



def internal_only(func):
    """Decorator to mark a function as deliberately not documented as it is not intended for public use."""
    func.__internal_only__ = True
    return func


internal_only(internal_only)



from neurosym.python_dsl.python_def_use_mask.defining_statement_handler.defining_statement_handler import (
    DefiningStatementHandler,
)

from .defining_statement_handler.defining_construct_handler import (
    DefiningConstructHandler,
)
from .extra_var import ExtraVar, canonicalized_python_name_as_leaf
from .handler import (
    ConstructHandler,
    DefaultHandler,
    Handler,
    HandlerPuller,
    default_handler,
)
from .mask import DefUseChainPreorderMask, DefUseMaskConfiguration
from .names import GLOBAL_REGEX, NAME_REGEX, match_either_name_or_global
from .ordering import PythonNodeOrdering, python_ordering_dictionary
from .special_case_symbol_predicate import SpecialCaseSymbolPredicate
from .target_handler import TargetHandler, create_target_handler



from abc import ABC, abstractmethod
from typing import Callable, List, Tuple

from neurosym.python_dsl.python_ast_tools import fields_for_node

from .names import match_either_name_or_global
from .special_case_symbol_predicate import SpecialCaseSymbolPredicate


class Handler(ABC):
    """
    Corresponds to a given node in the ``PythonAST``. Keeps track of a list of defined
    production indices, in order of definition, along with a reference to the mask, and
    the configuration.

    The handler is responsible for determining whether a given context is defining or not,
    and for computing the mask for a given set of names.

    The handler is also responsible for determining the handler for a given child, and for
    performing tasks related to the child when it is exited.

    :param mask: The ``DefUsePreorderMask`` that is being used, this is a circular reference
        used for convenience.
    :param defined_production_idxs: The list of defined production indices, in order of
        definition.
    :param config: The configuration for the mask.
    """

    def __init__(self, mask, defined_production_idxs, config):
        assert isinstance(defined_production_idxs, list)
        self.mask = mask
        self.defined_production_idxs = defined_production_idxs
        self.config = config

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple["Handler", Callable[[], None]]:
        """
        When a child is entered, this method is called to determine the handler for the child.
        This method can also perform tasks related to entering the child. By default, this
        method returns the default handler for the child, and does nothing else.

        This method can be undone by calling the returned function.

        :param position: The position in the s-expression.
        :param symbol: The symbol of the child (index into a grammar's symbols list).
            Note: this can include variables, e.g., const-&x:0-Name,
            but it can also include production symbols like Assign~S
            or other leaves like const-i2~Const.

        :return: ``(child_handler, undo)``: The handler for the child and a function
            that can be called to undo the changes made by this function.
        """
        return (
            default_handler(
                position,
                symbol,
                self.mask,
                self.currently_defined_indices(),
                self.config,
            ),
            lambda: None,
        )

    def on_child_exit(
        self, position: int, symbol: int, child: "Handler"
    ) -> Callable[[], None]:
        """
        When a child is exited, this method is called to perform tasks related to
        processing the information from the child. By default, this method does nothing.

        :param position: The position in the s-expression.
        :param symbol: The symbol of the child. See ``on_child_enter`` for more information.
        :param child: The handler for the child.

        :return: A function that can be called to undo the changes made by this function.
        """
        del position, symbol, child
        return lambda: None

    def currently_defined_indices(self) -> list[int]:
        """
        Returns the list of currently defined symbols, in order of definition. This is
        important because we need to know which symbols are currently defined for some
        purposes, e.g., when using imperative de bruijn indices.
        """
        assert isinstance(self.defined_production_idxs, list)
        return self.defined_production_idxs

    @abstractmethod
    def is_defining(self, position: int) -> bool:
        """
        Returns whether the context at the given position is defining.

        E.g., for an Assign node, the left-hand side is defining, and
        the right-hand side is not. This is important because for
        defining contexts, we do not need to use a previously
        defined variable, so new variables will not be masked out.

        :param position: The position in the s-expression.
        """

    def currently_defined_names(self):
        """
        Return the set of currently defined names. Note that this
        isn't the set of production symbols like ``const-&x:0-Name``,
        but rather a set of names like ``x``.
        """
        names = set()
        for symbol in self.currently_defined_indices():
            sym = self.mask.id_to_name(symbol)
            mat = match_either_name_or_global(sym)
            if not mat:
                continue
            names.add(mat.group("name"))
        return names

    def target_child(
        self, position: int, symbol: int
    ) -> Tuple["Handler", Callable[[], None]]:
        """
        Return a handler for a target. This is used when we are processing a target,
        e.g., the left-hand side of an assignment. By default, this method returns the
        default handler for the child, and does nothing else.

        :param position: The position in the s-expression.
        :param symbol: The symbol of the child. See ``on_child_enter`` for more information.
        """
        # pylint: disable=cyclic-import
        from .target_handler import create_target_handler

        return (
            create_target_handler(
                position,
                symbol,
                self.mask,
                self.currently_defined_indices(),
                self.config,
            ),
            lambda: None,
        )

    def _matches(self, names, symbol_id, idx_to_name, special_case_predicates):
        """
        Whether or not the symbol matches the names.
        """

        for pred in special_case_predicates:
            if pred.applies(symbol_id):
                return pred.compute(symbol_id, names)
        if idx_to_name[symbol_id] is None:
            return True
        return idx_to_name[symbol_id] in names

    def compute_mask(
        self,
        position: int,
        symbols: List[int],
        idx_to_name: List[str],
        special_case_predicates: List[SpecialCaseSymbolPredicate],
    ):
        """
        Compute the mask for the given names. If the context is defining,
        then all symbols are valid. Otherwise, only the symbols that
        match the names are valid.

        :param position: The position in the s-expression.
        :param symbols: The list of symbols to compute the mask for.
        :param idx_to_name: The list of names for the symbols.
        :param special_case_predicates: The list of special case predicates to use.
            A special case predicate is a way to extend the behavior of the handler
            by providing a custom predicate for a given symbol as to whether
            it should be considered valid or not.

        :return: A list of booleans, where ``True`` indicates that the symbol is valid.
        """
        assert isinstance(special_case_predicates, list) and all(
            isinstance(x, SpecialCaseSymbolPredicate) for x in special_case_predicates
        )
        if self.is_defining(position):
            return [True] * len(symbols)
        names = set(self.currently_defined_names())
        return [
            self._matches(names, symbol, idx_to_name, special_case_predicates)
            for symbol in symbols
        ]


class ConstructHandler(Handler):
    """
    Base class for Handlers that correspond to a specific construct in the AST.

    :field name: The name of the construct. This must be the name of a subclass of
        ``ast.AST``. This is used to determine the child fields of the construct.
    """

    # must be overridden in subclasses, represents the name of the construct
    name: str = None

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        assert isinstance(self.name, str)
        self.child_fields = {
            field: i for i, field in enumerate(fields_for_node(self.name))
        }


class DefaultHandler(Handler):
    """
    The default handler for a given symbol. This handler is used
    when a construct does nothing interesting with symbols. Effectively, just a
    passthrough handler.
    """

    def is_defining(self, position: int) -> bool:
        return False


def default_handler(
    position: int, symbol: int, mask, defined_production_idxs, config
) -> Handler:
    """
    Compute the default handler for a given symbol. This chooses the type
    of the handler based on the head symbol.

    :param position: The position in the s-expression.
    :param symbol: The symbol of the construct.
    :param mask: The mask being used.
    :param defined_production_idxs: The list of currently defined production indices.
    :param config: The configuration for the mask.
    """
    # pylint: disable=cyclic-import
    from .defining_statement_handler import defining_statement_handlers

    assert isinstance(defined_production_idxs, list)

    symbol = mask.id_to_name(symbol)
    pulled = config.pull_handler(position, symbol, mask, defined_production_idxs)
    if pulled is not None:
        return pulled

    return defining_statement_handlers().get(symbol, DefaultHandler)(
        mask, defined_production_idxs, config
    )


class HandlerPuller(ABC):
    """
    Represents a class that can pull a handler for a given symbol. This is
    used to extend the behavior of the DefUsePreorderMask by providing
    a space where hooks can be added to change the behavior of the mask
    on newly defined symbols.
    """

    @abstractmethod
    def pull_handler(
        self,
        position: int,
        symbol: int,
        mask,
        defined_production_idxs,
        config,
        handler_fn: Callable,
    ) -> Handler:
        """
        Pull a handler for the given symbol, position, etc.

        :param position: The position in the s-expression.
        :param symbol: The symbol of the construct.
        :param mask: The mask being used.
        :param defined_production_idxs: The list of currently defined production indices.
        :param config: The configuration for the mask.
        :param handler_fn: The function to call to get the default handler at the given location.
            This is provided because the default handler is not always the ``default_handler``
            function's result, i.e., when in a target context.

        :return: The handler to use for the given symbol.
        """
